#!/usr/bin/env python3
"""
WLAN Scanner
Advanced wireless network discovery and analysis tool.
Parameters:
    - interface: Wireless interface to use
    - duration: Scan duration in seconds
    - passive: Use passive scanning mode
Requires: scapy package and admin/root privileges
"""
from scapy.all import *
import argparse
import logging
from datetime import datetime
import json
import threading
import queue
from collections import defaultdict
import time
import os
import signal
import subprocess
import math

class WLANScanner:
    def __init__(self, interface, duration=60, passive=True):
        self.interface = interface
        self.duration = duration
        self.passive = passive
        self.stop_event = threading.Event()
        self.networks = defaultdict(dict)
        self.clients = defaultdict(set)
        self.hidden_ssids = set()
        self.channel_stats = defaultdict(int)
        self.setup_logging()

    def setup_logging(self):
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s'
        )
        self.logger = logging.getLogger(__name__)

    def enable_monitor_mode(self):
        """Enable monitor mode on wireless interface"""
        try:
            self.logger.info(f"Enabling monitor mode on {self.interface}")
            subprocess.run(['airmon-ng', 'check', 'kill'], check=True)
            subprocess.run(['airmon-ng', 'start', self.interface], check=True)
            return True
        except Exception as e:
            self.logger.error(f"Error enabling monitor mode: {str(e)}")
            return False

    def disable_monitor_mode(self):
        """Disable monitor mode on wireless interface"""
        try:
            subprocess.run(['airmon-ng', 'stop', self.interface], check=True)
            subprocess.run(['service', 'NetworkManager', 'start'], check=True)
            self.logger.info("Monitor mode disabled")
        except Exception as e:
            self.logger.error(f"Error disabling monitor mode: {str(e)}")

    def set_channel(self, channel):
        """Set wireless interface to specific channel"""
        try:
            subprocess.run(
                ['iwconfig', self.interface, 'channel', str(channel)],
                check=True
            )
        except Exception as e:
            self.logger.error(f"Error setting channel: {str(e)}")

    def calculate_signal_quality(self, rssi):
        """Calculate signal quality percentage from RSSI"""
        if rssi >= -50:
            quality = 100
        elif rssi <= -100:
            quality = 0
        else:
            quality = 2 * (rssi + 100)
        return quality

    def analyze_network_security(self, network):
        """Analyze security features of a network"""
        security = {
            'encryption': [],
            'key_mgmt': [],
            'auth': [],
            'cipher': [],
            'score': 0,
            'issues': []
        }

        # Check encryption
        if network.get('encryption'):
            for enc in network['encryption']:
                if enc == 'WEP':
                    security['encryption'].append('WEP')
                    security['score'] -= 50
                    security['issues'].append(
                        'WEP encryption is highly vulnerable'
                    )
                elif enc == 'WPA':
                    security['encryption'].append('WPA')
                    security['score'] += 30
                elif enc == 'WPA2':
                    security['encryption'].append('WPA2')
                    security['score'] += 40
                elif enc == 'WPA3':
                    security['encryption'].append('WPA3')
                    security['score'] += 50
        else:
            security['encryption'].append('OPEN')
            security['score'] -= 100
            security['issues'].append('Network is open and unencrypted')

        # Check for hidden SSID
        if not network.get('ssid'):
            security['score'] += 5
            security['issues'].append(
                'Hidden SSID provides minimal security benefit'
            )

        # Check client isolation
        if len(network.get('clients', [])) > 1:
            security['score'] -= 10
            security['issues'].append(
                'Multiple clients detected - verify client isolation'
            )

        # Normalize score
        security['score'] = max(0, min(100, security['score']))
        return security

    def packet_handler(self, pkt):
        """Process captured packets"""
        if self.stop_event.is_set():
            return

        try:
            # Process Beacon frames
            if pkt.haslayer(Dot11Beacon):
                bssid = pkt[Dot11].addr2
                ssid = pkt[Dot11Elt].info.decode('utf-8', errors='ignore')
                channel = int(ord(pkt[Dot11Elt:3].info))
                
                try:
                    rssi = -(256-ord(pkt.notdecoded[-4:-3]))
                except:
                    rssi = -100
                
                capability = pkt.sprintf("{Dot11Beacon:%Dot11Beacon.cap%}")
                encryption = []
                
                # Determine encryption type
                if 'privacy' in capability:
                    encryption.append('WEP')
                
                # Check for WPA/WPA2
                p = pkt[Dot11Elt]
                while isinstance(p, Dot11Elt):
                    if p.ID == 48:  # RSN (WPA2)
                        encryption.append('WPA2')
                    elif p.ID == 221 and p.info.startswith(b'\x00P\xf2\x01\x01\x00'):
                        encryption.append('WPA')
                    p = p.payload

                if not encryption:
                    encryption = ['OPEN']

                # Update network information
                self.networks[bssid].update({
                    'ssid': ssid,
                    'channel': channel,
                    'rssi': rssi,
                    'quality': self.calculate_signal_quality(rssi),
                    'encryption': encryption,
                    'capability': capability,
                    'beacons': self.networks[bssid].get('beacons', 0) + 1,
                    'last_seen': datetime.now().isoformat()
                })
                
                self.channel_stats[channel] += 1
                
                if not ssid:
                    self.hidden_ssids.add(bssid)

            # Process Probe Requests
            elif pkt.haslayer(Dot11ProbeReq):
                if pkt.haslayer(Dot11):
                    client_mac = pkt[Dot11].addr2
                    if client_mac not in ['00:00:00:00:00:00', 'ff:ff:ff:ff:ff:ff']:
                        if pkt.haslayer(Dot11Elt) and pkt[Dot11Elt].info:
                            ssid = pkt[Dot11Elt].info.decode('utf-8', errors='ignore')
                            if ssid:
                                self.clients[client_mac].add(ssid)

            # Process Data frames
            elif pkt.haslayer(Dot11):
                ds = pkt.FCfield & 0x3
                if ds == 1:
                    bssid = pkt[Dot11].addr1
                    client = pkt[Dot11].addr2
                elif ds == 2:
                    bssid = pkt[Dot11].addr2
                    client = pkt[Dot11].addr1
                else:
                    return

                if client not in ['00:00:00:00:00:00', 'ff:ff:ff:ff:ff:ff']:
                    if bssid in self.networks:
                        self.networks[bssid]['clients'] = \
                            self.networks[bssid].get('clients', set())
                        self.networks[bssid]['clients'].add(client)

        except Exception as e:
            self.logger.error(f"Error processing packet: {str(e)}")

    def channel_hopper(self):
        """Hop through WiFi channels"""
        while not self.stop_event.is_set():
            try:
                for channel in range(1, 14):
                    if self.stop_event.is_set():
                        break
                    self.set_channel(channel)
                    time.sleep(0.5)
            except Exception as e:
                self.logger.error(f"Error in channel hopper: {str(e)}")
                time.sleep(0.5)

    def save_results(self, output_file):
        # Convert sets to lists for JSON serialization
        networks_json = {}
        for bssid, network in self.networks.items():
            network_copy = network.copy()
            if 'clients' in network_copy:
                network_copy['clients'] = list(network_copy['clients'])
            network_copy['security_analysis'] = \
                self.analyze_network_security(network_copy)
            networks_json[bssid] = network_copy

        clients_json = {}
        for client, ssids in self.clients.items():
            clients_json[client] = list(ssids)

        results = {
            'timestamp': datetime.now().isoformat(),
            'interface': self.interface,
            'scan_duration': self.duration,
            'networks': networks_json,
            'clients': clients_json,
            'channel_statistics': dict(self.channel_stats),
            'hidden_networks': list(self.hidden_ssids)
        }

        try:
            with open(output_file, 'w') as f:
                json.dump(results, f, indent=4)
            self.logger.info(f"Results saved to {output_file}")
        except Exception as e:
            self.logger.error(f"Error saving results: {str(e)}")

    def print_summary(self):
        self.logger.info("\nWLAN Scan Summary:")
        self.logger.info(f"Networks discovered: {len(self.networks)}")
        self.logger.info(f"Clients discovered: {len(self.clients)}")
        self.logger.info(f"Hidden networks: {len(self.hidden_ssids)}")
        
        # Channel utilization
        self.logger.info("\nChannel Utilization:")
        for channel, count in sorted(self.channel_stats.items()):
            self.logger.info(f"Channel {channel}: {count} networks")
        
        # Network details
        self.logger.info("\nNetworks by Signal Strength:")
        sorted_networks = sorted(
            self.networks.items(),
            key=lambda x: x[1].get('quality', 0),
            reverse=True
        )
        
        for bssid, network in sorted_networks:
            security = self.analyze_network_security(network)
            self.logger.info(
                f"\nBSSID: {bssid}\n"
                f"SSID: {network.get('ssid', 'Hidden')}\n"
                f"Channel: {network.get('channel')}\n"
                f"Signal: {network.get('rssi')} dBm "
                f"({network.get('quality')}%)\n"
                f"Encryption: {', '.join(network.get('encryption', []))}\n"
                f"Security Score: {security['score']}/100\n"
                f"Clients: {len(network.get('clients', set()))}"
            )
            
            if security['issues']:
                self.logger.info("Security Issues:")
                for issue in security['issues']:
                    self.logger.info(f"- {issue}")

    def start_scan(self, output_file=None):
        if not self.enable_monitor_mode():
            return

        try:
            self.logger.info(f"Starting WLAN scan on {self.interface}")
            self.logger.info(f"Scan duration: {self.duration} seconds")
            self.logger.info(f"Mode: {'Passive' if self.passive else 'Active'}")
            
            # Start channel hopper
            hopper = threading.Thread(target=self.channel_hopper)
            hopper.daemon = True
            hopper.start()

            # Start packet capture
            sniff(
                iface=self.interface,
                prn=self.packet_handler,
                timeout=self.duration,
                store=0
            )

        except KeyboardInterrupt:
            self.logger.info("Scan interrupted by user")
        except Exception as e:
            self.logger.error(f"Error during scan: {str(e)}")
        finally:
            self.stop_event.set()
            if output_file:
                self.save_results(output_file)
            self.print_summary()
            self.disable_monitor_mode()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="WLAN Scanner")
    parser.add_argument("interface", help="Wireless interface to use")
    parser.add_argument("--duration", type=int, default=60,
                      help="Scan duration in seconds")
    parser.add_argument("--passive", action="store_true",
                      help="Use passive scanning mode")
    parser.add_argument("--output", help="Output file for scan results")
    args = parser.parse_args()

    scanner = WLANScanner(args.interface, args.duration, args.passive)
    scanner.start_scan(args.output)
